package com.EndCaptcha;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.util.Date;
import java.util.Hashtable;
import java.util.Dictionary;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Base EndCaptcha API client.
 *
 */
abstract public class Client
{
    final static public String API_VERSION = "EC/Java v1.1";
    final static public int SOFTWARE_VENDOR_ID = 0;

    final static public int DEFAULT_TIMEOUT = 60;
    final static public int DEFAULT_TOKEN_TIMEOUT = 120;
    final static public int[] POLLS_INTERVAL = {1, 1, 2, 3, 2, 2, 3, 2, 2};
    final static public int DFLT_POLL_INTERVAL = 3;
    final static public int LEN_POLLS_INTERVAL = POLLS_INTERVAL.length;


    /**
     * Client verbosity flag.
     *
     * When it's set to true, the client will dump API calls for debug purpose.
     */
    public boolean isVerbose = false;


    protected String _username = "";
    protected String _password = "";


    protected void log(String call, String msg)
    {
        if (this.isVerbose) {
            System.out.println((System.currentTimeMillis() / 1000) + " " +
                               call + (null != msg ? ": " + msg : ""));
        }
    }

    protected void log(String call)
    {
        this.log(call, null);
    }

    protected JSONObject getCredentials()
    {
        try {
            return new JSONObject().put("username", this._username).put("password", this._password);
        } catch (JSONException e) {
            return new JSONObject();
        }
    }

    protected byte[] load(InputStream st)
        throws IOException
    {
        int n = 0, offset = 0;
        byte[] img = new byte[0];
        while (true) {
            try {
                n = st.available();
            } catch (IOException e) {
                n = 0;
            }
            if (0 < n) {
                if (offset + n > img.length) {
                    img = java.util.Arrays.copyOf(img, img.length + n);
                }
                offset += st.read(img, offset, n);
            } else {
                break;
            }
        }
        return img;
    }

    protected byte[] load(File f)
        throws IOException, FileNotFoundException
    {
        InputStream st = new FileInputStream(f);
        try {
            return this.load(st);
        } catch (FileNotFoundException e) {
            throw e;
        } catch (IOException e) {
            throw e;
        } finally {
            st.close();
        }
    }

    public byte[] load(String fn)
        throws IOException, FileNotFoundException
    {
        return this.load(new File(fn));
    }

    /**
     * @param username DBC account username
     * @param password DBC account password
     */
    public Client(String username, String password)
    {
        this._username = username;
        this._password = password;
    }

    /**
     * Fetches user balance (in US cents).
     *
     * @return Dictionary user balance
     */
    public Dictionary getBalance()
        throws IOException, com.EndCaptcha.Exception
    {
        return this.getBalance();
    }


    /**
     * Uploads a CAPTCHA to the service.
     *
     * @param img CAPTCHA image byte vector
     * @return CAPTCHA object on success, null otherwise
     */
    abstract public Dictionary upload(byte[] img, String challenge, int type, byte[] banner, String banner_text, String grid)
        throws IOException, com.EndCaptcha.Exception;

    // abstract public Captcha upload(byte[] img, String challenge, int type, byte[] banner, String banner_text)
    //     throws IOException, com.EndCaptcha.Exception;
    //
    // abstract public Captcha upload(byte[] img, int type, byte[] banner, String banner_text)
    //     throws IOException, com.EndCaptcha.Exception;

    abstract public Dictionary upload(byte[] img)
        throws IOException, com.EndCaptcha.Exception;

    /**
     * @see com.EndCaptcha.Client#upload
     * @param st CAPTCHA image stream
     */
    public Dictionary upload(InputStream st)
        throws IOException, com.EndCaptcha.Exception
    {
        return this.upload(this.load(st));
    }

    /**
     * @see com.EndCaptcha.Client#upload
     * @param f CAPTCHA image file
     */
    public Dictionary upload(File f)
        throws IOException, FileNotFoundException, com.EndCaptcha.Exception
    {
        return this.upload(this.load(f));
    }

    /**
     * @see com.EndCaptcha.Client#upload
     * @param fn CAPTCHA image file name
     */
    public Dictionary upload(String fn)
        throws IOException, FileNotFoundException, com.EndCaptcha.Exception
    {
        return this.upload(this.load(fn));
    }

    // /**
    //  * @see com.EndCaptcha.Client#upload
    //  * @param type CAPTCHA type
    //  * @param proxy User proxy
    //  * @param proxytype User proxy type
    //  * @param sitekey Site Google Key
    //  * @param pageurl Site url
    //  */
    // public Captcha upload(int type, String proxy, String proxytype, String sitekey, String pageurl)
    //     throws IOException, com.EndCaptcha.Exception
    //     {
    //         JSONObject json = new JSONObject();
    //             try{
    //             json.put("proxy",proxy);
    //             json.put("proxytype",proxytype);
    //             json.put("googlekey",sitekey);
    //             json.put("pageurl",pageurl);
    //         } catch (JSONException e) {
    //             //System.out.println(e);
    //         }
    //         return this.upload(type,json);
    //     }
    abstract public Dictionary upload(int type,JSONObject json)
        throws IOException, com.EndCaptcha.Exception;


    /**
     * Fetches an uploaded CAPTCHA details.
     *
     * @param id CAPTCHA ID
     * @return CAPTCHA object if found, null otherwise
     */
    abstract public Dictionary get_captcha(int id)
        throws IOException, com.EndCaptcha.Exception;

    /**
     * Fetches an uploaded CAPTCHA text.
     *
     * @param id CAPTCHA ID
     * @return CAPTCHA text if solved, null otherwise
     */
    public String get_text(int id)
        throws IOException, com.EndCaptcha.Exception
    {
        Dictionary captcha = new Hashtable();

        captcha = this.get_captcha(id);

        if (null != captcha.get("text")) {
          return captcha.get("text").toString();
        } else {
          return null;
        }
    }

    /**
    * Reports an incorrectly solved CAPTCHA
    *
    * @param CAPTCHA
    * @return Dictionary
    */
    abstract public Dictionary report(byte[] img)
        throws IOException, com.EndCaptcha.Exception;
 
    /**
    * Reports an incorrectly solved CAPTCHA
    *
    * @param CAPTCHA
    * @return Dictionary
    */
    abstract public Dictionary report(String fn, int type)
        throws IOException, com.EndCaptcha.Exception;


    /**
    * Reports an incorrectly solved CAPTCHA
    *
    * @param CAPTCHA
    * @return Dictionary
    */
    public Dictionary report(String fn)
        throws IOException, com.EndCaptcha.Exception
    {
        Boolean fnf;

        try {
            byte[] img = this.load(fn);
            return this.report(img);
        } catch (java.io.FileNotFoundException e) {
            fnf = true;
        }

        if (fnf) {
            String pattern_hash = "^[0-9a-f]{40}$";
            String pattern_int = "^[0-9]+$";
            
            Pattern reh = Pattern.compile(pattern_hash);
            Matcher reh_match = reh.matcher(fn);
            Pattern rei = Pattern.compile(pattern_int);
            Matcher rei_match = rei.matcher(fn);
            
            if (reh_match.find()) {
                return this.report(fn, 1);
            }
            if (rei_match.find()) {
                return this.report(fn, 2);
            }
        }
        
        throw new java.io.FileNotFoundException("image (No such file or directory)");
    }

    /**
    * Reports an incorrectly solved CAPTCHA
    *
    * @param CAPTCHA
    * @return Dictionary
    */
    public Dictionary report(int fn)
        throws IOException, com.EndCaptcha.Exception
    {
        return this.report(Integer.toString(fn), 2);
    }

    /**
    * Reports an incorrectly solved CAPTCHA
    *
    * @param CAPTCHA
    * @return Dictionary
    */
    public Dictionary report(File f)
        throws IOException, com.EndCaptcha.Exception
    {
        return this.report(this.load(f));
    }

    /**
    * Reports an incorrectly solved CAPTCHA
    *
    * @param CAPTCHA
    * @return Dictionary
    */
    public Dictionary report(InputStream st)
        throws IOException, com.EndCaptcha.Exception
    {
        return this.report(this.load(st));
    }

    /**
     * Tries to solve a CAPTCHA by uploading it and polling for its status
     * and text with arbitrary timeout.
     *
     * @param img CAPTCHA image byte vector
     * @param timeout Solving timeout (in seconds)
     * @return CAPTCHA object if uploaded and correctly solved, null otherwise
     */
    public Dictionary decode(byte[] img, String challenge, int type, byte[] banner, String banner_text, String grid, int timeout)
        throws IOException, com.EndCaptcha.Exception, InterruptedException
    {
        long deadline = System.currentTimeMillis() + (0 < timeout ? timeout : Client.DEFAULT_TIMEOUT) * 1000;
        Dictionary captcha = this.upload(img, challenge, type, banner, banner_text, grid);

        if (null != captcha) {
            int intvl_idx = 0;
            int intvl = 0;
            int[] results = {0, 0};
            int cid = -1;

            if (null != captcha.get("unsolved_yet")) {
                cid = Integer.parseInt((String)captcha.get("unsolved_yet"));
            }

            while (deadline > System.currentTimeMillis() && null == captcha.get("text") && cid > 0) {
                results = Client.getPollInterval(intvl_idx);
                intvl = results[0];
                intvl_idx = results[1];
                Thread.sleep(intvl * 1000);
                captcha = this.get_captcha(cid);
            }
            if (null != captcha.get("text")) {
                return captcha;
            }
        }

        return null;
    }

    public Dictionary decode(byte[] img, String challenge, int type, byte[] banner, String banner_text, int timeout)
        throws IOException, com.EndCaptcha.Exception, InterruptedException
    {
        return this.decode(img, challenge, type, banner, banner_text, "", 0);
    }




    /**
     * @see com.EndCaptcha.Client#decode(byte[], int)
     */
    public Dictionary decode(byte[] img)
        throws IOException, com.EndCaptcha.Exception, InterruptedException
    {
        return this.decode(img, "", 0,null, "", 0);
    }

    /**
     * @see com.EndCaptcha.Client#decode(byte[], int)
     */
    public Dictionary decode(byte[] img, int timeout)
        throws IOException, com.EndCaptcha.Exception, InterruptedException
    {
        return this.decode(img, "", 0,null, "", timeout);
    }

    public Dictionary decode(File f, int timeout)
        throws IOException, FileNotFoundException, com.EndCaptcha.Exception, InterruptedException
    {
        return this.decode(this.load(f), timeout);
    }

    /**
     * @see com.EndCaptcha.Client#decode(File, int)
     */
    public Dictionary decode(File f)
        throws IOException, FileNotFoundException, com.EndCaptcha.Exception, InterruptedException
    {
        return this.decode(f, 0);
    }


    public Dictionary decode(String fn, int timeout)
        throws IOException, FileNotFoundException, com.EndCaptcha.Exception, InterruptedException
    {
        return this.decode(this.load(fn), timeout);
    }

    public Dictionary decode(int type, JSONObject json)
    throws IOException, FileNotFoundException, com.EndCaptcha.Exception, InterruptedException
    {

        return this.decode(type, json, 0);
    }

    public Dictionary decode(int type, JSONObject json, int timeout)
        throws IOException, com.EndCaptcha.Exception, InterruptedException
    {
        long deadline = System.currentTimeMillis() + (0 < timeout ? timeout : Client.DEFAULT_TOKEN_TIMEOUT) * 1000;
        Dictionary captcha = this.upload(type, json);

        if (null != captcha) {
            int intvl_idx = 0;
            int intvl = 0;
            int[] results = {0, 0};
            int cid = -1;

            if (null != captcha.get("unsolved_yet")) {
                cid = Integer.parseInt((String)captcha.get("unsolved_yet"));
            }

            while (deadline > System.currentTimeMillis() && null == captcha.get("text") && cid > 0) {
                results = Client.getPollInterval(intvl_idx);

                intvl = results[0];
                intvl_idx = results[1];
                Thread.sleep(intvl * 1000);
                captcha = this.get_captcha(cid);
            }

            if (null != captcha.get("text")) {

                if (cid > 0) {
                    captcha.put("captcha_id", cid);
                }

                return captcha;
            }

            return captcha;
        }
        
        return null;
    }

    /**
     * @param int index of POLLS_INTERVAL
     * @return position of POLLS_INTERVAL or default
     * polling time if index out of range
     */
    public static int[] getPollInterval(int idx)
    {
        int intvl = 0;

        if (Client.LEN_POLLS_INTERVAL > idx ) {
            intvl = Client.POLLS_INTERVAL[idx];
        }

        else {
            intvl = Client.DFLT_POLL_INTERVAL;
        }

        return new int[]{intvl, ++idx};
    }
}
